@@ -1,4 +1,4 @@
-# Copyright 2024 IQM
+# Copyright 2024-2025 IQM
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,6 +15,6 @@
 """Utilities for working with Cirq objects."""
 
 
-def cirq_to_cpc():
+def cirq_to_cpc():  # noqa: ANN201
     """Convert a Cirq Circuit to an IQM CPC Circuit."""
     raise NotImplementedError("Cirq to IQM conversion is not yet implemented.")
